
# We import the usual package:
using JuMP
using StochDynamicProgramming
using CPLEX

using PyPlot

const PLOT = false

# Read optimization scenarios:
De = readcsv("demandelec_opt.csv")
Dt = readcsv("demandtherm_opt.csv")

# Read assessment scenarios:
De_assess = readcsv("demandelec_assess.csv")
Dt_assess = readcsv("demandtherm_assess.csv");

if PLOT
    subplot(2, 1, 1)
    plot(De, c="darkblue", lw=.5);
    xticks(0:4*4:96, 0:4:24);
    grid()
    subplot(2, 1, 2)
    plot(Dt, c="darkblue", lw=.5);
    xticks(0:4*4:96, 0:4:24);
    grid();
end

################################################################################
################################################################################
#### CONSTANT DEFINITION
################################################################################
################################################################################
# Set physical variables of system:
# Timestep
DT = .25
NTIME = Int(24/DT)
N_NOISES = size(De, 2)

CHP_POWER = 7
CHP_YIELD = .93
SHARE_ELEC = .13
P_CHP_ELEC = CHP_POWER * CHP_YIELD * SHARE_ELEC
P_CHP_THERM = CHP_POWER * CHP_YIELD * (1- SHARE_ELEC)

# Battery settings:
ALPHA_B =1
BETA_B = 1
DELTA_B_MAX = .6

# Boiler discharge rate:
ALPHA_H = 1.
BETA_H = 1.

# Set prices:
P_GAS = .06
P_CHP = P_GAS*CHP_POWER
P_BURNER = P_GAS/.9
P_ELEC = 0.09
P_ELEC_PEAK = 0.15
P_INJECTION = .0
P_TH = .12

# Set initial configuration:
B_0 = .6
H_0 = 20


# Generate prices other day:
priceElec = ones(NTIME) * P_ELEC
for nd=0:0
    priceElec[8*4+nd*96:22*4+nd*96] = P_ELEC_PEAK
end

PLOT && plot(priceElec);

################################################################################
################################################################################
#### MODEL DEFINITION
################################################################################
################################################################################
function dynamic(t, x, u, w)
    return [ALPHA_B * x[1] - DT*BETA_B*u[2],
            min(max(ALPHA_H * x[2] + DT*BETA_H*(P_CHP_THERM*u[1] + u[3] - w[2]), 19), 30)]
end

function cost(t, x, u, w)
    output_flow = min(w[2], (x[2] - 19.)/(BETA_H*DT) + u[3] + P_CHP_THERM*u[1])

    return P_BURNER*u[3] + P_CHP*u[1] +
           priceElec[t]* max(0, (w[1] - u[2] - P_CHP_ELEC * u[1]))  + P_TH * (w[2] - output_flow)
end

finalcost(x) = 0.

constraint(t, x, u, w) = true;

################################################################################
################################################################################
#### PROBABILISTIC MODELLING
################################################################################
################################################################################
"""Fit marginal probability laws with optimization scenarios."""
function generate_probability_laws(de, dt)
    aleas = zeros(NTIME, N_NOISES, 2)
    aleas[:, :, 1] = de
    aleas[:, :, 2] = dt

    laws = Vector{NoiseLaw}(NTIME)

    # uniform probabilities:
    pp = 1/N_NOISES*ones(N_NOISES)

    for t=1:NTIME
        support = reshape(aleas[t, :, :], 2, 30)
        laws[t] = NoiseLaw(support, pp)
    end

    return laws
end

################################################################################
################################################################################
#### EVALUATION & SIMULATION OF POLICY
################################################################################
################################################################################
function simulate(scenarios, policy)
    nassess = size(scenarios, 2)
    ntime = size(scenarios, 1)

    x0 = [.6, 20.]
    stocks = zeros(ntime, nassess, 2)
    controls = zeros(ntime-1, nassess, 3)
    costs = zeros(nassess)

    # Set first value of stocks equal to x0:
    for i in 1:nassess
        stocks[1, i, :] = x0
    end

    for t=1:ntime-1
        for k = 1:nassess
            state_t = stocks[t, k, :]
            aleas = scenarios[t, k, :]

            uc = policy(t, state_t)

            xf = dynamic(t, state_t, uc, aleas)

            stocks[t+1, k, :] = xf
            controls[t, k, :] = uc
            costs[k] += cost(t, state_t, uc, aleas)
        end
    end

    return costs, stocks, controls
end

scenarios = zeros(NTIME, 30, 2)
scenarios[:, :, 1] = De_assess
scenarios[:, :, 2] = Dt_assess

"""Simple heuristic."""
function policy_heuristic(t, x)
    H_THRESH = 21.
    H_LIM = 19.
    fa = 0
    y = 0
    #= if fa > 0. && (x[2] > 29.) =#
    #=     y = 0 =#
    #= end =#
    if x[2] < H_THRESH
        y = 1
    end
    if x[2] < H_LIM
        fa = 12.
        y = 1
    end

    return [y, 0, fa]
end

# Launch assessment:
cheur, xheur, uheur = simulate(scenarios, policy_heuristic)
println("Cost heuristic: ", .25*mean(cheur))

################################################################################
################################################################################
#### MODEL PREDICTIVE CONTROL
################################################################################
################################################################################
De_prevision = mean(De_assess, 2)
Dt_prevision = mean(Dt_assess, 2);
previsions = hcat(De_prevision, Dt_prevision);

# shrinking horizon MPC:
function policy_mpc(t::Int64, x0::Array{Float64})
    tf = NTIME - t

    m = Model()

    # Define constraints other variables:
    @variable(m,  19           <= h[1:(tf+1)]  <= 30)
    @variable(m,  0.5          <= b[1:(tf+1)]  <= 3)
    @variable(m, -DELTA_B_MAX  <= Fb[1:(tf)]   <= DELTA_B_MAX)
    @variable(m,  0.           <= Fa[1:(tf)]   <= 5.5)
    @variable(m,  0 <= Fh[1:(tf)] <= 20)
    @variable(m,  0 <= Yt[1:(tf)] <= 1)
    @variable(m,  Z1[1:(tf)] >= 0)


    # Set optimization objective:
    @objective(m, Min, sum{P_BURNER*Fa[i] + P_CHP*Yt[i] + Z1[i] + P_TH*Fh[i], i = 1:tf})

    for i in 1:tf
        @constraint(m, b[i+1] - ALPHA_B * b[i] + DT*BETA_B*Fb[i] == 0)
        @constraint(m, h[i+1] - ALPHA_H * h[i] -
                       DT*BETA_H*(Yt[i]*P_CHP_THERM + Fa[i] - previsions[(i+t -1), 2] + Fh[i]) == 0)

        @constraint(m, Z1[i] >=   (previsions[(i+t -1), 1] - Fb[i] - Yt[i]*P_CHP_ELEC)*priceElec[t+i-1])
        @constraint(m, Z1[i] >=  -(previsions[(i+t -1), 1] - Fb[i] - Yt[i]*P_CHP_ELEC)*P_INJECTION)
    end

    # Add initial constraints:
    @constraint(m, b[1] == x0[1])
    @constraint(m, h[1] == x0[2])

    # Solve:
    status = solve(m)
    return [getvalue(Yt)[1]; getvalue(Fb)[1]; getvalue(Fa)[1]]
end


cmpc, xmpc, umpc = simulate(scenarios, policy_mpc);
println("Cost MPC: ", .25mean(cmpc))


################################################################################
################################################################################
#### STOCHASTIC DYNAMIC PROGRAMMING
################################################################################
################################################################################
# Instantiate model:
x0 = [.6, 20]
aleas = generate_probability_laws(De_assess, Dt_assess)

xbounds = [(.5, 3), (19, 30)]
ubounds = [(0, 1), (-DELTA_B_MAX, DELTA_B_MAX), (0, 5.5), (0, 15)]

sdpmodel = StochDynProgModel(NTIME, xbounds, ubounds[1:3], x0,
                             cost, finalcost, dynamic, constraint, aleas);

stateSteps = [.2, 1]           # discretization step of the state
controlSteps = [.5, .1, .2]      # discretization step of the control
infoStruct = "HD"               # noise at tim t is known before taking the decision at time t
paramSDP = SDPparameters(sdpmodel, stateSteps, controlSteps, infoStruct);

# Solve problem with dynamic programming:
Vs = solve_DP(sdpmodel, paramSDP, 1)
value_sdp = StochDynamicProgramming.get_bellman_value(sdpmodel, paramSDP, Vs)
println("Value obtained by SDP: " , round(value_sdp,4))


csdp, xsdp, usdp = sdp_forward_simulation(sdpmodel, paramSDP, scenarios, Vs, true);
println("Cost SDP: ", .25*mean(csdp))


if PLOT
    figure()
    subplot(2, 1, 1)
    plot(statessdp[:, 1:30, 1], c="k", lw=.2);
    ylabel("Battery SOC (kWh)")
    grid()
    xticks(0:4*4:96, 0:4:24);
    subplot(2, 1, 2)
    plot(statessdp[:, 1:30, 2], c="k", lw=.2);
    grid()
    ylabel("Tank level (kWh)")
    xlabel("hours")
    xticks(0:4*4:96, 0:4:24);

    figure()
    subplot(2, 1, 1)
    plot(stockssddp[:, 1:30, 1], c="k", lw=.2);
    ylabel("Battery SOC (kWh)")
    grid()
    xticks(0:4*4:96, 0:4:24);
    subplot(2, 1, 2)
    plot(stockssddp[:, 1:30, 2], c="k", lw=.2);
    grid()
    ylabel("Tank level (kWh)")
    xlabel("hours")
    xticks(0:4*4:96, 0:4:24);
end

